# Create.pm
# implements 'create' command
#
# v0.21, 2003-06-29
# Copyright (c) Michel Messerschmidt 2003
#
package Avtc::Create;
use strict;
use Exporter;
use Cwd;
use File::Basename;
use File::Copy;
use File::Path;

@Avtc::Create::ISA = qw(Exporter);
@Avtc::Create::EXPORT = qw(createPacked);


# creates packed testbed contents from unpacked testbed
sub createPacked {
    my $ret = 0;
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    my $tb;
    LOOP_TB: foreach $tb (sort keys %::lists) {
        if (not exists($::lists{$tb}->{"packed"})) {
            # skip all unpacked testbeds
            next;
        }
        my $type = $::lists{$tb}->{"packed"}->{"type"};
        my $tbRoot = $::lists{$tb}->{"path"}->{$::configOs};
        my $refTb = $tb;
        ######### attention: uses 5.char !!!
        substr($refTb, 3, 2) = "__";
        my $refRoot = $::listsAll{$refTb}->{"path"}->{$::configOs};

        ::log("creating testbed $tb ($tbRoot) from testbed $refTb ($refRoot)".
              "\n", 1);

        my $tempDir =  $::TempBase . $::sep . $tb;
        # used to remember directories to change back to
        my $backDir = "";
        ::makeDir($tempDir) or do {
            ::log("ERROR creating $tempDir:  $!\n");
            ::log("      !!! Creation of $tb stopped !!!\n\n");
            $ret = 1;
            next LOOP_TB;
        };
        chdir($tempDir.$::sep) or do {
            ::log("ERROR: chdir $tempDir.$::sep failed:  $!\n");
            ::log("      !!! Creation of $tb stopped !!!\n\n");
            $ret = 1;
            next LOOP_TB;
        };
        if ($::debug) {
            ::log("  Current dir set to " . cwd() . "\n");
        }

        if ($type eq "P" or $type eq "E" or $type eq "S") {
            # create list with all directories that must be packed
            my @list = ();
            my $listFile = $::ListDir . $::sep . $::FilePreDirlist .
                          $::listsAll{$refTb}->{"id"} .".". $::FileExtLists;
            open(LIST, "< $listFile") or do {
                ::log("ERROR opening dirlist '$listFile'\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };
            binmode(LIST);
            my $tempLine;
            my $lastDir = "";
            while (defined($tempLine = <LIST>)) {
                # cut filename from path
                $tempLine =~ s/^(.*)(\\|\/)[^\\\/]+$/$1/;
                # skip lines with identical directories
                if ($tempLine eq $lastDir) { next; }
                $lastDir = $tempLine;
                push @list, $tempLine;
            }
            close(LIST);

            my $listRef = \@list;

            # create the testbed
            createStandardTb($tb, $tbRoot, $refRoot, $backDir, $listRef);

        }
        elsif ($type eq "Q") {
            my $password = $::lists{$tb}->{"packed"}->{"password"};
            my %prg = %{$::lists{$tb}->{"packed"}->{"prg"}};
            # create directory if needed
            if ($::debug) {
                ::log("  creating dir ".$tbRoot.$::sep."\n");
            }
            ::makeDir($tbRoot.$::sep) or do {
                ::log("ERROR creating $tbRoot.$::sep:  $!\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };
            # all archive are created with relative paths
            chdir($refRoot.$::sep) or do {
                ::log("ERROR: chdir $refRoot.$::sep failed:  $!\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };
            if ($::debug) {
                ::log("  Current dir set to " . cwd() . "\n");
            }


            # create the tar archive first
            # (because this is used for the compression only programs)
            my $pack = "TAR";
            my $hasTar = 0;
            my $tarFile;
            if (exists $prg{$pack}) {
                $hasTar = 1;
                # get real input filenames to replace "%i" in command
                my @in = ($prg{$pack}->{"wildcard"});
                $tarFile = $prg{$pack}->{"filename"};
                my $out = $tbRoot . $::sep . $tarFile;
                createArchive($prg{$pack}->{"add"}, $password, $out, @in)
                or do {
                    ::log("ERROR creating archives for program $pack\n", 1);
                };
                $ret = 1;
            }
            LOOP_PRG: foreach $pack (sort keys %::packList) {
                if ((not exists $prg{$pack}) or ($pack eq "TAR")) {
                    # this compression program is not used for this testbed
                    # (and TAR archives were created before this loop)
                    next LOOP_PRG;
                }
                # get real output filename to replace "%o" in command
                my $wildcard = $prg{$pack}->{"wildcard"};
                my $out = $tbRoot . $::sep . $prg{$pack}->{"filename"};
                # get real input filenames to replace "%i" in command
                my @in = ($wildcard);
                # oh no! special cases...
                if ($prg{$pack}->{"onlyComp"}) {
                    if (not $hasTar) {
                        ::log("skipping program $pack\n", 1);
                        next LOOP_PRG;
                    }
                    # this program can only compress single files
                    @in = ($tbRoot.$::sep.$tarFile);
                    # "onlyComp" AND "onlyRelative" needs some more lines
                    if ($prg{$pack}->{"onlyRelative"}) {
                        # use relative filenames for input and output
                        # remember current dir
                        $backDir = cwd();
                        chdir($tbRoot.$::sep) or do {
                            ::log("ERROR: chdir $tbRoot.$::sep failed:  $!\n");
                            ::log("ERROR archive not created for program ".
                                  $pack."\n");
                            $ret = 1;
                            next LOOP_PRG;
                        };
                        @in = ($tarFile);
                        $out = $prg{$pack}->{"filename"};
                        if ($::debug) {
                            ::log("  Current dir set to " . cwd() . "\n");
                        }
                    }
                }
                # no special instructions needed for this program
                createArchive($prg{$pack}->{"add"}, $password, $out, @in) or do {
                    ::log("ERROR creating archives for program $pack\n");
                    $ret = 1;
                };

                # change back to root of reference testbed
                if ($backDir) {
                    chdir($backDir.$::sep) or do {
                        ::log("ERROR: chdir $backDir.$::sep failed:  $!\n");
                        ::log("      !!! Creation of $tb stopped !!!\n\n");
                        $ret = 1;
                        next LOOP_TB;
                    };
                    if ($::debug) {
                        ::log("  Current dir set to " . cwd() . "\n");
                    }
                    $backDir = "";
                }
            } # end of LOOP_PRG

        }
        elsif ($type eq "R") {
            my $password = $::lists{$tb}->{"packed"}->{"password"};
            my $recDepth = $::lists{$tb}->{"packed"}->{"recursive"};
            my %prg = %{$::lists{$tb}->{"packed"}->{"prg"}};

            # create directory if needed
            if ($::debug) {
                ::log("  creating dir $tbRoot.$::sep\n");
            }
            ::makeDir($tbRoot.$::sep) or do {
                ::log("ERROR creating $tbRoot.$::sep:  $!\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };

            # create list with all directories that must be packed
            my @dirList = ();
            my @tbDirList = ();
            my $listFile = $::ListDir . $::sep . $::FilePreDirlist .
                          $::listsAll{$refTb}->{"id"} .".". $::FileExtLists;
            open(LIST, "< $listFile") or do {
                ::log("ERROR opening dirlist '$listFile'\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };
            binmode(LIST);
            my $tempLine;
            my $lastDir = "";
            while (defined($tempLine = <LIST>)) {
                # cut filename from path
                $tempLine =~ s/^(.*)(\\|\/)[^\\\/]+$/$1/;
                # skip lines with identical directories
                if ($tempLine eq $lastDir) { next; }
                $lastDir = $tempLine;
                push @dirList, $tempLine;
                # replace root directory from unpacked (reference)
                # testbed with packed testbed rootdir
                $tempLine =~ s/^\Q$refRoot\E(.*)$/$tbRoot$1/;
                # backslashes "\" in $tbRoot may be inserted doubled
                # (= uninterpreted "\\") here, so remove them
                $tempLine =~ s/\\\\/\\/g;
                push @tbDirList, $tempLine;
            }
            close(LIST);

            my $dirListRef = \@dirList;
            my $tbDirListRef = \@tbDirList;

            # create initial (non-recursive packed) archives
            createStandardTb($tb, $tbRoot, $refRoot, $backDir, $dirListRef);


            # all recursive compressions are done in the $::TempRoot directory
            # (to avoid errors with long paths/commandlines)
            chdir($::TempRoot.$::sep) or do {
                ::log("ERROR: chdir $::TempRoot.$::sep failed:  $!\n");
                ::log("      !!! Creation of $tb stopped !!!\n\n");
                $ret = 1;
                next LOOP_TB;
            };
            if ($::debug) {
                ::log("  Current dir set to " . cwd() . "\n");
            }

            my $rec;
            for ($rec = 0; $rec < $recDepth; $rec++) {
                if ($::debug) {
                    ::log("  Creating next recursion (depth $rec)...\n");
                }
                # create next recursion depth archives
                createRecTb($tb, $tbDirListRef) or do {
                    ::log("      !!! Creation of $tb stopped !!!\n\n");
                    $ret = 1;
                    next LOOP_TB;
                };
            }

        }
        else {
            ::log("ERROR: unknown packed testbed type '$type'\n");
            $ret = 1;
        }
    }
    return $ret;
}


#syntax: processTb($testbed, $dirlistReference);
# This function expects to have the current dir set to $::TempRoot
# It will NOT work correctly otherwise
sub createRecTb {
    my $tb = $_[0];
    my @list = @{$_[1]};

    my $password = $::lists{$tb}->{"packed"}->{"password"};
    my %prg = %{$::lists{$tb}->{"packed"}->{"prg"}};
    my $ok = 0;   # to check return values from I/O operations

    my $pack;
    LOOP_PRG: foreach $pack (sort keys %::packList) {
        if (not exists $prg{$pack}) {
            # this compression program is not used for this testbed
            next LOOP_PRG;
        }
        if ($::debug) {
            ::log("  creating archives for program $pack\n");
            if ($prg{$pack}->{"onlyComp"}) {
                ::log("    program is 'compress only'\n");
            }
        }
        my $addCmd = $prg{$pack}->{"add"};
        my $archFile = $prg{$pack}->{"filename"};

        my $tbDir;
        LOOP_DIR: foreach $tbDir (@list) {
            if ($::debug) {
                ::log("    processing dir '$tbDir'\n");
            }
            # additional check for non-existing directories
            if (not -e $tbDir) {
                ::log("ERROR non-existing directory '$tbDir':  $!\n");
                return 0;
            }
            if ($prg{$pack}->{"onlyComp"} and not -e $tbDir.$::sep.$archFile) {
                ::log("ERROR non-existing directory '".
                      $tbDir.$::sep.$archFile."'\n");
                return 0;
            }

            # create random name for temporary archive
            my $tempArchive = "temp_" . $pack;
            # oh no! special cases...
            if ($pack eq "BH_") {
                # program uses file extension to select archive type
                # the archive file is renamed to $archFile after creation
                $tempArchive .= ".BH";
                if ($::debug) {
                    ::log("      appending 'BH' extension to archive ".
                          "filename for program BH_\n");
                }
            }

            if ($prg{$pack}->{"onlyComp"}) {
                # this program can only compress single files
                # Therefore $archFile denotes a directory instead of the
                # archive file and we have each original file inside it
                # (but compressed)
                my $destDir = $tbDir . $::sep . $archFile;
                my @compList = ();
                my $tempName;
                opendir(DIR, $destDir) or do {
                    ::log("ERROR reading $destDir: $!\n");
                    next LOOP_DIR;
                };
                while (defined($tempName = readdir(DIR))) {
                    # skip directory entries
                    if ($tempName eq "." or $tempName eq "..") { next; }
                    if ($::debug > 1) {
                        ::log("      found file '$tempName' in dir '".
                              $destDir."'\n");
                    }
                    push @compList, $tempName;
                }
                closedir(DIR);

                my $compFile;
                foreach $compFile (@compList) {
                    createRecArchive($addCmd, $password, $tempArchive,
                                     $compFile, $destDir) or do {
                        ::log("ERROR archive not created for program $pack\n");
                        next LOOP_DIR;
                    };
                }

            } else {
                createRecArchive($addCmd, $password, $tempArchive,
                                 $archFile, $tbDir) or do {
                    ::log("ERROR archive not created for program $pack\n");
                    next LOOP_DIR;
                };
            }
        } # end of LOOP_DIR
    } # end of LOOP_PRG
    return 1;
}


sub createRecArchive {
    my $add = $_[0];         # command for archive creation
    my $password = $_[1];    # password for archive creation
    my $tempArchive = $_[2]; # temporary archive filename in $::TempRoot
    my $archFile = $_[3];    # archive file (found in $tbDir) that should be packed once more
                             # with the same compression program
    my $tbDir = $_[4];       # testbed directory to use

    # source and destination file for the compressed archive
    my $destFile = $tbDir . $::sep . $archFile;
    # get real input filenames to replace "%i" in command
    my @in =($archFile);
    # get real output filename to replace "%o" in command
    my $out = $tempArchive;

    if ($::debug) {
        ::log("      copying '".$destFile."' to '".
              $::TempRoot.$::sep.$archFile."'\n");
    }
    copy($destFile, $::TempRoot.$::sep.$archFile)
    or do {
        ::log("ERROR copying '".$destFile."' to '".
              $::TempRoot.$::sep.$archFile."': $!\n");
        return 0;
    };

    createArchive($add, $password, $out, @in)
    or do {
        ::log("ERROR creating archive\n");
    };
    # handle file extensions that may be appended by a
    # compression program
    my $f;
    my @fl;
    my $found = 0;
    opendir(DIR, $::TempRoot.$::sep) or do {
        ::log("ERROR reading $::TempRoot.$::sep: $!\n");
        # dont abort here, maybe we have luck and it wasn't
        # required to read the filenames
        # if it is serious, then it will fail soon
    };
    while (defined($f = readdir(DIR))) {
        push @fl, $f;
    }
    closedir(DIR);
    foreach $f (@fl) {
        if ($f =~ /^\Q$tempArchive\E$/i) {
            $found = 1;
            last;
        }
    }
    if (not $found) {
        foreach $f (@fl) {
            if ($f =~ /^\Q$tempArchive\E\..*$/i) {
                if ($::debug) {
                    ::log("      found temporary archive '$f'\n");
                }
                $tempArchive = $f;
                last;
            }
        }
    }

    unlink($destFile) or do {
        ::log("ERROR deleting old '$destFile': $!\n");
    };

    if ($::debug) {
        ::log("      copying '".$::TempRoot.$::sep.$tempArchive."' to '".
              $destFile."'\n");
    }
    my $ok = copy($::TempRoot.$::sep.$tempArchive, $destFile);
    unlink($tempArchive) or do {
        ::log("ERROR deleting $tempArchive: $!\n");
    };
    unlink($archFile) or do {
        ::log("ERROR deleting $archFile: $!\n");
    };
    if (not $ok) {
        ::log("ERROR copying '".$::TempRoot.$::sep.$tempArchive."' to '".
              $destFile."': $!\n");
        return 0;
    };
    return 1;
}


#syntax: processTb($tb, $refTb, $refRoot, $tbRoot, $backDir, $dirListReference);
sub createStandardTb {
    my $tb = $_[0];
    my $tbRoot = $_[1];
    my $refRoot = $_[2];
    my $backDir = $_[3];
    my @list = @{$_[4]};

    my $password = $::lists{$tb}->{"packed"}->{"password"};
    my %prg = %{$::lists{$tb}->{"packed"}->{"prg"}};

    my $pack;
    LOOP_PRG: foreach $pack (sort keys %::packList) {
        if (not exists $prg{$pack}) {
            # this compression program is not used for this testbed
            next LOOP_PRG;
        }
        my ($refDir, $dir);
        my $wildcard = $prg{$pack}->{"wildcard"};
        LOOP_DIR: foreach $refDir (@list) {
            if ($backDir) {
                chdir($backDir.$::sep) or do {
                    ::log("ERROR: chdir $backDir.$::sep failed:  $!\n");
                    ::log("      !!! Creation of $tb stopped !!!\n\n");
                    next LOOP_TB;
                };
                if ($::debug) {
                    ::log("  Current dir set to " . cwd() . "\n");
                }
                $backDir = "";
            }
            # replace root directory from unpacked (reference)
            # testbed with packed testbed rootdir
            $dir = $refDir;
            $dir =~ s/^\Q$refRoot\E(.*)$/$tbRoot$1/;
            # backslashes "\" in $tbRoot may be inserted doubled
            # (= uninterpreted "\\") here, so remove them
            $dir =~ s/\\\\/\\/g;
            if ($::debug > 1) {
                ::log("    tbRoot='$tbRoot'  - refRoot='$refRoot'\n");
                ::log("    oldDir='$refDir'  - newDir='$dir'\n");
            }
            # create directory if it's not there
            if (not -e $dir) {
                if ($::debug) {
                    ::log("  creating dir $dir\n");
                }
                ::makeDir($dir) or do {
                    ::log("ERROR creating $dir:  $!\n");
                    ::log("      !!! Creation of $tb stopped !!!\n\n");
                    next LOOP_TB;
                };
            }

            # get real output filename to replace "%o" in command
            my $archFile = $prg{$pack}->{"filename"};

            # get real input filenames to replace "%i" in command
            my @in = ($refDir . $::sep . $wildcard);

            my $out = $dir . $::sep . $archFile;
            # oh no! special cases...
            if ($pack eq "BH_" and $out !~ /\.BH$/i) {
                # program uses file extension to select archive type
                # the archive file is renamed to $archFile after creation
                $out .= ".BH";
            }
            if ($prg{$pack}->{"onlyRelative"}) {
                # use relative filenames for input:
                # remember current dir
                $backDir = cwd();
                chdir($refDir.$::sep) or do {
                    ::log("ERROR: chdir $refDir.$::sep failed:  $!\n");
                    ::log("      !!! Creation of $tb stopped !!!\n\n");
                    next LOOP_TB;
                };
                @in = ($wildcard);
                if ($::debug) {
                    ::log("  Current dir set to " . cwd() . "\n");
                }
            }

            if ($prg{$pack}->{"onlyComp"}) {
                # this program can only compress single files
                # create a directory instead of the archive file
                # and store compressed samples in it
                my $archDir = $out;
                if ($::debug) {
                    ::log("    creating special dir $archDir\n");
                }
                ::makeDir($archDir) or do {
                    ::log("ERROR creating $archDir:  $!\n");
                    next LOOP_PRG;
                };
                my $srcDir = $refDir;
                opendir(TEMP, $srcDir) or do {
                    ::log("ERROR reading $srcDir: $!\n");
                    next LOOP_PRG;
                };
                my $tempName;
                while (defined($tempName = readdir(TEMP))) {
                    # skip directory entries
                    if ($tempName eq "." or $tempName eq "..") { next; }
                    my $compOut = $archDir . $::sep . $tempName;
                    my @compIn = ($srcDir . $::sep . $tempName);
                    if ($prg{$pack}->{"onlyRelative"}) {
                        # use relative filenames for input
                        @compIn = ($tempName);
                    }
                    createArchive($prg{$pack}->{"add"}, $password,
                                  $compOut, @compIn) or do {
                        ::log("ERROR creating archives for program ".
                              $pack."\n");
                    };
                }
                closedir(TEMP);
            } else {
                # no special instructions needed for this program
                createArchive($prg{$pack}->{"add"}, $password, $out, @in) or do {
                    ::log("ERROR creating archives for program ".
                          $pack."\n");
                };

                # handle filename alterations by compression programs
                my $f;
                my @fl;
                my $found = 0;
                opendir(DIR, $dir) or do {
                    ::log("ERROR reading $dir: $!\n");
                    # dont abort here, maybe we have luck and it wasn't
                    # required to read the filenames
                    # if it is serious, then it will fail soon
                };
                while (defined($f = readdir(DIR))) {
                    push @fl, $f;
                }
                closedir(DIR);
                foreach $f (@fl) {
                    if ($f =~ /^\Q$archFile\E$/i) {
                        $found = 1;
                        last;
                    }
                }
                if (not $found) {
                    my $archHead = $archFile;
                    $archHead =~ s/^(.*)\..*$/$1/;
                    foreach $f (@fl) {
                        if (($f =~ /^\Q$archFile\E\..*$/i) or
                            ($f =~ /^\Q$archHead\E.*\..*$/i)) {
                            # the first regex should catch things like rar1.rar.01
                            # the second regex should catch things like rar1.exe
                            if ($::debug) {
                                ::log("DEBUG:  found temporary archive '".$f.
                                      "'\n");
                            }
                            move($dir.$::sep.$f, $dir.$::sep.$archFile) or do {
                                ::log("ERROR renaming '$f' to '$archFile': $!\n");
                            };
                            last;
                        }
                    }
                }

            }
        } # end of LOOP_DIR
    } # end of LOOP_PRG
}


# create a single archive using the commandline given
# commandline flags %o, %i and %p are replaced by passed arguments
sub createArchive {
    my $extCall = shift @_;
    my $password = shift @_;
    my $out = shift @_;
    my @in = @_;
    if ($::debug) {
        ::log("      command:  $extCall\n");
    }
    # replace password placeholder with password string
    $extCall =~ s/\%p/$password/g;
    # replace output file placeholder with output filename
    $extCall =~ s/\%o/$out/g;
    # replace input file placeholder with input filenames list
    my $in = "";
    foreach (@in) {
        $in .= $_ . " ";
    }
    chop($in);
    $extCall =~ s/\%i/$in/g;

    my $errlog = 0;
    if ($extCall !~ /\>/) {
        $extCall .= " >".$::TempBase.$::sep."extcall.tmp 2>&1";
        $errlog = 1;
    }

    if ($::debug) {
        ::log("      executing '$extCall'\n");
    }
    my $error = system($extCall);
    if ($::debug) {
        ::log("      ...external program terminated with exitcode $error\n");
    }
    if ($error) {
        ::log("ERROR in external command: exit code=$error, error=$!\n");
        if ($errlog) {
            open(TMP, "< ".$::TempBase.$::sep."extcall.tmp") and do {
                ::log("External commands output:\n", 1);
                my $line;
                while (defined($line=<TMP>)) {
                    ::log($line, 2);
                }
            };
            close(TMP);
            unlink($::TempBase.$::sep."extcall.tmp");
        } else {
            ::log("  Couldn't log external commands output\n");
        }
        return 0;
    }
    return 1;
}


# return a true value
1;