# Eval_chk.pm
# implements consistency checks for all evaluation steps
#
# v1.02, 2004-03-29
# Copyright (c) Michel Messerschmidt 2002-2004
#
package Avtc::Eval_chk;
use strict;
use Exporter;
use File::Copy;
use File::Basename;
use Cwd;
use IO::File;

@Avtc::Eval_chk::ISA = qw(Exporter);
@Avtc::Eval_chk::EXPORT = qw(checkTbCommon checkTb checkPackedAll checkList
                             checkListFP);

# do common checks that are possible on all testbeds
sub checkTbCommon {
    my $os = $_[0];
    my $tb = $_[1];
    my $prod = $_[2];
    my $checkFile = $_[3];
    my $dirList = $_[4];
    my ($lSbo, $lNrm, $lD_f, $lO_f, $lA_f, $lMis, $lGbf, $lS_f, $lIno,
        $lPsl, $lS_g, $lEv1, $lRe1, $lDfp, $lMfp, $lFpEv1, $lFpGbf,
        $lFpA_f, $lFiles, $lFdf, $lFof, $lFpFdf, $lFFpof
        );
    # get check data
    ($lFiles, undef) = ::wc($dirList);
    ($lSbo, undef) = ::wc($prod . "." . $::FileExtScannedOK);
    ($lNrm, undef) = ::wc($prod . "." . $::FileExtNorm);
    ($lD_f, undef) = ::wc($prod . "." . $::FileExtDetFiles);
    ($lO_f, undef) = ::wc($prod . "." . $::FileExtOKFiles);
    ($lA_f, undef) = ::wc($prod . "." . $::FileExtAnoFiles);
    ($lMis, undef) = ::wc($prod . "." . $::FileExtMissedFiles);
    ($lGbf, undef) = ::wc($prod . "." . $::FileExtGarbageFiles);
    ($lS_f, undef) = ::wc($prod . "." . $::FileExtScanFiles);
    ($lIno, undef) = ::wc($prod . "." . $::FileExtInfAndOK);
    ($lPsl, undef) = ::wc($prod . "." . $::FileExtPostscan);
    ($lS_g, undef) = ::wc($prod . "." . $::FileExtGarbage);
    ($lEv1, undef) = ::wc($prod . "." . $::FileExtEval);
    $lFdf = $::data{"DF_Filtered"}->{$os}->{$tb}->{$prod};
    $lFof = $::data{"OF_Filtered"}->{$os}->{$tb}->{$prod};

    my $prodFP = "";
    if (exists $::lists{$tb}->{"fp"}) {
        $prodFP = $prod . "_" . $::FileIdFP;
        ($lDfp, undef) = ::wc($prod . "." . $::FileExtDetFPFiles);
        ($lMfp, undef) = ::wc($prod . "." . $::FileExtMissedFPFiles);
        ($lRe1, undef) = ::wc($prod . "." . $::FileExtRem);
        ($lFpGbf, undef) = ::wc($prodFP . "." . $::FileExtGarbageFiles);
        ($lFpA_f, undef) = ::wc($prodFP . "." . $::FileExtAnoFiles);
        ($lFpEv1, undef) = ::wc($prodFP . "." . $::FileExtEval);
        $lFpFdf = $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP};
        $lFFpof = $::data{"OF_Filtered"}->{$os}->{$tb}->{$prodFP};
    } else {
        # set all line numbers to 0, since these files don't exist.
        # Any other value would cause wrong results below !!!
        $lDfp = 0;
        $lMfp = 0;
        $lRe1 = 0;
        $lFpGbf = 0;
        $lFpA_f = 0;
        $lFpEv1 = 0;
        $lFpFdf = 0;
        $lFFpof = 0;
    }

    my $packSpace = "   ";
    my $packProg = "";
    if (exists $::lists{$tb}->{"packed"}) {
        $packSpace = "";
        $packProg = substr($prod, 3);
        if ($packProg eq $::FileExtArchiveFiles) {
            $packProg = "archives";
        }
    }

    # write check data
    my $openError = 0;
    open(CHECK, ">> $checkFile") or do {
        ::log("ERROR opening $checkFile: $!\n", 2);
        ::log("You must check this manually!\n", 5);
        $openError = 1;
    };
    binmode(CHECK);

    if (not $openError) {
        if ($packProg ne "") {
            # write an additional header for packed testbeds
            print(CHECK $::EOL);
            print(CHECK "  check data for ",$packProg, $::EOL);
            print(CHECK $::EOL, "                                          ",
                  "   lines   check result", $::EOL);
        }
        print(CHECK "---- common checks for all testbeds --------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   normalized logfile (%s.%s)       %s: %7d",
               $prod, $::FileExtNorm, $packSpace, $lNrm);
        if (($lNrm - $lFdf - $lFpFdf) == ($lD_f + $lDfp)) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " - filtered duplicate lines              : %7d%s",
               $lFdf, $::EOL);
        if ($prodFP ne "") {
            # testbeds with false positives have some additional files
            printf(CHECK " - filtered duplicate fp lines           : %7d%s",
                   $lFpFdf, $::EOL);
        }
        printf(CHECK " = files detected by scanner (%s.%s)%s: %7d%s",
               $prod, $::FileExtDetFiles, $packSpace, $lD_f, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + detected fp files (%s.%s)        %s: %7d%s",
                   $prod, $::FileExtDetFPFiles, $packSpace, $lDfp, $::EOL);
        }

        print(CHECK "============================================",
                    "=============================", $::EOL);
        printf(CHECK "   scanned but reported as OK(%s.%s)%s: %7d",
               $prod, $::FileExtScannedOK, $packSpace, $lSbo);
        if (($lSbo - $lFof) == $lO_f) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " - filtered duplicate lines              : %7d%s",
               $lFof, $::EOL);
        printf(CHECK " = files scanned but ok (%s.%s)     %s: %7d%s",
               $prod, $::FileExtOKFiles, $packSpace, $lO_f, $::EOL);
        print(CHECK "============================================",
                    "=============================", $::EOL);
        printf(CHECK "   Files in Testbed (from dirlist)       : %7d",
               $lFiles);
        if ($lFiles == ($lA_f + $lFpA_f + $lMis + $lMfp - $lGbf
                        - $lFpGbf + $lD_f + $lDfp)) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " = anomaly files (%s.%s)            %s: %7d%s",
               $prod, $::FileExtAnoFiles, $packSpace, $lA_f, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + anomaly fp files (%s.%s)       %s: %7d%s",
                   $prodFP, $::FileExtAnoFiles, $packSpace, $lFpA_f,
                   $::EOL);
        }
        printf(CHECK " + missed samples (%s.%s)           %s: %7d%s",
               $prod, $::FileExtMissedFiles, $packSpace, $lMis, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + missed fp samples (%s.%s)        %s: %7d%s",
                   $prod, $::FileExtMissedFPFiles, $packSpace, $lMfp, $::EOL);
        }
        printf(CHECK " - garbage files in .%s (%s.%s)    %s: %7d",
               $::FileExtNorm, $prod, $::FileExtGarbageFiles,
               $packSpace, $lGbf);
        if ($lGbf != 0) { print(CHECK "    Key ERROR!!!"); }
        print(CHECK $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " - garbage fp files .%s (%s.%s)  %s: %7d",
                   $::FileExtNorm, $prodFP, $::FileExtGarbageFiles,
                   $packSpace, $lFpGbf);
            if ($lFpGbf != 0) { print(CHECK "    Key ERROR!!!"); }
            print(CHECK $::EOL);
        }
        printf(CHECK " + files detected by scanner (%s.%s)%s: %7d%s",
               $prod, $::FileExtDetFiles, $packSpace, $lD_f, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + detected fp files (%s.%s)        %s: %7d%s",
                   $prod, $::FileExtDetFPFiles, $packSpace, $lDfp, $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   files detected by scanner (%s.%s)%s: %7d%s",
               $prod, $::FileExtDetFiles, $packSpace, $lD_f, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + detected fp files (%s.%s)        %s: %7d%s",
                   $prod, $::FileExtDetFPFiles, $packSpace, $lDfp, $::EOL);
        }
        printf(CHECK " + files scanned but ok (%s.%s)     %s: %7d%s",
               $prod, $::FileExtOKFiles, $packSpace, $lO_f, $::EOL);
        printf(CHECK " - files infected AND ok (%s.%s)    %s: %7d",
               $prod, $::FileExtInfAndOK, $packSpace, $lIno);
        if ($lIno != 0) { print(CHECK "    check this"); }
        print(CHECK $::EOL);
        printf(CHECK " = all files scanned (%s.%s)        %s: %7d",
               $prod, $::FileExtScanFiles, $packSpace, $lS_f);
        if (($lD_f + $lDfp + $lO_f - $lIno) == $lS_f) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   all files scanned (%s.%s)        %s: %7d%s",
               $prod, $::FileExtScanFiles, $packSpace, $lS_f, $::EOL);
        printf(CHECK " + files in postscan list (%s.%s)   %s: %7d",
               $prod, $::FileExtPostscan, $packSpace, $lPsl);
        if ($lPsl != 0) {
            if (exists $::lists{$tb}->{"packed"}) {
                # packed testbeds may need an extra warning here
                if ($lNrm == 0 and $lO_f == 0) {
                    print(CHECK " ignore this!", $::EOL,
                  "      (if the normalized log is empty, ",
                                "this postscan list is irrelevant)");
                } else {
                    print(CHECK "    postscan!");
                }
            } else {
                print(CHECK "    postscan!");
            }
        }
        print(CHECK $::EOL);
        printf(CHECK " - all evaluated garbage (%s.%s)    %s: %7d",
               $prod, $::FileExtGarbage, $packSpace, $lS_g);
        if ($lS_g != 0) { print(CHECK "    Key ERROR!!!"); }
        print(CHECK $::EOL);
        printf(CHECK " = Files in Testbed (from dirlist)       : %7d",
               $lFiles);
        if (($lS_f + $lPsl - $lS_g) == $lFiles) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        print(CHECK "============================================",
                    "=============================", $::EOL);
        printf(CHECK "   normalized logfile (%s.%s)       %s: %7d",
               $prod, $::FileExtNorm, $packSpace, $lNrm);
        if ($lNrm == ($lFpEv1 + $lEv1 + $lRe1)) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " = malware lines evaluated  (%s.%s) %s: %7d%s",
               $prod, $::FileExtEval, $packSpace, $lEv1, $::EOL);
        if ($prodFP ne "") {
            printf(CHECK " + false positive lines (%s.%s)   %s: %7d%s",
                   $prodFP, $::FileExtEval, $packSpace, $lFpEv1, $::EOL);
            printf(CHECK " + remaining lines from log (%s.%s) %s: %7d",
                   $prod, $::FileExtRem, $packSpace, $lRe1);
            if ($lRe1 != 0) { print(CHECK "    check this"); }
            print(CHECK $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL, $::EOL);
        close(CHECK);
    }
}


# check testbed evaluation
sub checkTb {
    my $os = $_[0];
    my $tb = $_[1];
    my $prod = $_[2];
    my $checkFile = $_[3];
    my $dirList = $_[4];
    my ($lFul, $lRes, $lSbo, $lAus, $lAur, $lNru, $lNrm, $lE_f, $lFiles);
    # test if file with redundant evaluation lines exists
    my $aur = $prod . "." . $::FileExtAuswertRedundant;
    my $hasAur = 0;
    if (-e $aur) {
        $hasAur = 1;
        ($lAur, undef) = ::wc($aur);
    } else {
        $lAur = 0;
    }
    # get check data
    ($lFiles, undef) = ::wc($::ListDir . $::sep . $::FilePreDirlist .
                              $::lists{$tb}->{"id"} . "." . $::FileExtLists);
    ($lFiles, undef) = ::wc($dirList);
    ($lFul, undef) = ::wc($prod . "." . $::FileExtReport);
    ($lRes, undef) = ::wc($prod . "." . $::FileExtReportRem);
    ($lSbo, undef) = ::wc($prod . "." . $::FileExtScannedOK);
    ($lAus, undef) = ::wc($prod . "." . $::FileExtAuswert);
    ($lNru, undef) = ::wc($prod . "." . $::FileExtNormUnsorted);
    ($lE_f, undef) = ::wc($prod . "." . $::FileExtExcludedFiles);
    ($lNrm, undef) = ::wc($prod.".".$::FileExtNorm);

    # write check data
    my $openError = 0;
    open(CHECK, ">> $checkFile") or do {
        ::log("ERROR opening $checkFile: $!\n", 2);
        ::log("You must check this manually!\n", 5);
        $openError = 1;
    };
    binmode(CHECK);

    if (not $openError) {
        print(CHECK $::EOL, "                                          ",
              "   lines   check result", $::EOL);
        print(CHECK "==== checks for scanner-key output =========",
                    "=============================", $::EOL);
        printf(CHECK "   original reportfile (%s.%s)         : %7d",
                 $prod, $::FileExtReport, $lFul);
        if (($lAus + $lRes) == $lFul) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " = non-evaluated lines (%s.%s)         : %7d%s",
               $prod, $::FileExtReportRem, $lRes, $::EOL);
        printf(CHECK " + evaluated lines (%s.%s)             : %7d%s",
               $prod, $::FileExtAuswert, $lAus, $::EOL);
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   evaluated lines (%s.%s)             : %7d",
               $prod, $::FileExtAuswert, $lAus);
        if (($lNru + $lSbo + $lAur) == $lAus) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        if ($hasAur) {
            printf(CHECK " = redundant evaluated lines (%s.%s)   : %7d%s",
                   $prod, $::FileExtAuswertRedundant, $lAur, $::EOL);
            printf(CHECK " + scanned but reported as OK (%s.%s)  : %7d%s",
           $prod, $::FileExtScannedOK, $lSbo, $::EOL);
        } else {
            printf(CHECK " = scanned but reported as OK (%s.%s)  : %7d%s",
           $prod, $::FileExtScannedOK, $lSbo, $::EOL);
        }
        printf(CHECK " + unsorted normalized logfile (%s.%s) : %7d%s",
               $prod, $::FileExtNormUnsorted, $lNru, $::EOL);
        print(CHECK "---- checks for normalized log evaluation --",
                    "-----------------------------", $::EOL);
        printf(CHECK "   unsorted normalized logfile (%s.%s) : %7d",
               $prod, $::FileExtNormUnsorted, $lNru);
        if (($lNru -$lE_f) == $lNrm) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " - excluded samples (%s.%s)            : %7d%s",
               $prod, $::FileExtExcludedFiles, $lE_f, $::EOL);
        printf(CHECK " = final normalized logfile (%s.%s)    : %7d%s",
               $prod, $::FileExtNorm, $lNrm, $::EOL);
        close(CHECK);
    }
}


# packed testbeds need to combine all single checks in one file
# some additional checks for the whole testbed evaluation are required too
sub checkPackedAll {
    my $os = $_[0];
    my $tb = $_[1];
    my $prod = $_[2];
    my $checkFile = $_[3];
    my ($lFul, $lRes, $lSbo, $lAus, $lAur, $lNru, $lNrm, $lRen, $lRps,
        $lAdd, $lAdo, $lPsl, $lE_f);
    my (%lNrmP, %lSboP);
    my $lpackNrmSum = 0;
    my $lpackSboSum = 0;

    # test if file with redundant evaluation lines exists
    my $aur = $prod . "." . $::FileExtAuswertRedundant;
    my $hasAur = 0;
    if (-e $aur) {
        $hasAur = 1;
        ($lAur, undef) = ::wc($aur);
    } else {
        $lAur = 0;
    }
    # get check data
    ($lFul, undef) = ::wc($prod . "." . $::FileExtReport);
    ($lRes, undef) = ::wc($prod . "." . $::FileExtReportRem);
    ($lSbo, undef) = ::wc($prod . "." . $::FileExtScannedOK);
    ($lAus, undef) = ::wc($prod . "." . $::FileExtAuswert);
    ($lNru, undef) = ::wc($prod . "." . $::FileExtNormUnsorted);
    ($lE_f, undef) = ::wc($prod . "." . $::FileExtExcludedFiles);
    ($lNrm, undef) = ::wc($prod . "." . $::FileExtNorm);
    ($lRen, undef) = ::wc($prod . "." . $::FileExtNormRemPacked);
    ($lRps, undef) = ::wc($prod . "." . $::FileExtScanOKRemPacked);
    ($lAdd, undef) = ::wc($prod . "." . $::FileExtPackedAdd);
    ($lAdo, undef) = ::wc($prod . "." . $::FileExtPackedOKAdd);
    ($lPsl, undef) = ::wc($prod . "." . $::FileExtPostscan);

    my $pack;
    foreach $pack (sort keys %{$::listsAll{$tb}->{"packed"}->{"prg"}}) {
        ($lNrmP{$pack}, undef) = ::wc($prod.$pack.".".$::FileExtNorm);
        $lpackNrmSum += $lNrmP{$pack};
        ($lSboP{$pack}, undef) = ::wc($prod.$pack.".".$::FileExtScannedOK);
        $lpackSboSum += $lSboP{$pack};
    }
    $pack = $::FileExtArchiveFiles;
    ($lNrmP{$pack}, undef) = ::wc($prod . $pack . "." . $::FileExtNorm);
    $lpackNrmSum += $lNrmP{$pack};
    ($lSboP{$pack}, undef) = ::wc($prod . $pack . "." . $::FileExtScannedOK);
    $lpackSboSum += $lSboP{$pack};

    # write check data
    my $openError = 0;
    open(CHECK, ">> $checkFile") or do {
        ::log("ERROR opening $checkFile: $!\n", 2);
        ::log("You must check this manually!\n", 5);
        $openError = 1;
    };
    binmode(CHECK);

    if (not $openError) {
        print(CHECK $::EOL, "                                          ",
                    "   lines   check result", $::EOL);
        print(CHECK "============================================",
                    "=============================", $::EOL);
        printf(CHECK "   postscan list info (%s.%s)          : %7d",
                     $prod, $::FileExtPostscan, $lPsl);
        if ($lPsl != 0) { print(CHECK "    postscan!"); }
        print(CHECK $::EOL);
        if ($lPsl != 0 and $lSbo > 0) {
            print(CHECK "    This postscanlist may contain scanned files. ");
            printf(CHECK "Check against %s.%s !%s",
                         $prod, $::FileExtScannedOK, $::EOL);
        }
        print(CHECK "==== checks for scanner-key output =========",
                    "=============================", $::EOL);
        printf(CHECK "   original reportfile (%s.%s)         : %7d",
                     $prod, $::FileExtReport, $lFul);
        if (($lAus + $lRes) == $lFul) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " = non-evaluated lines (%s.%s)         : %7d%s",
                     $prod, $::FileExtReportRem, $lRes, $::EOL);
        printf(CHECK " + evaluated lines (%s.%s)             : %7d%s",
                     $prod, $::FileExtAuswert, $lAus, $::EOL);
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   evaluated lines (%s.%s)             : %7d",
                     $prod, $::FileExtAuswert, $lAus);
        if (($lNru + $lSbo + $lAur) == $lAus) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        if ($hasAur) {
            printf(CHECK " = redundant evaluated lines (%s.%s)   :  %6d %s",
                   $prod, $::FileExtAuswertRedundant, $lAur, $::EOL);
            printf(CHECK " + scanned but reported as OK (%s.%s)  :  %6d %s",
                   $prod, $::FileExtScannedOK, $lSbo, $::EOL);
        } else {
            printf(CHECK " = scanned but reported as OK (%s.%s)  :  %6d %s",
                   $prod, $::FileExtScannedOK, $lSbo, $::EOL);
        }
        printf(CHECK " + unsorted normalized logfile (%s.%s) : %7d%s",
                     $prod, $::FileExtNormUnsorted, $lNru, $::EOL);
        print(CHECK "---- checks for normalized log evaluation --",
                    "-----------------------------", $::EOL);
        printf(CHECK "   unsorted normalized logfile (%s.%s) : %7d",
                     $prod, $::FileExtNormUnsorted, $lNru);
        if (($lNru -$lE_f) == $lNrm) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        printf(CHECK " - excluded samples (%s.%s)            : %7d%s",
                     $prod, $::FileExtExcludedFiles, $lE_f, $::EOL);
        printf(CHECK " = final normalized logfile (%s.%s)    : %7d%s",
                     $prod, $::FileExtNorm, $lNrm, $::EOL);
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   sorted normalized logfile (%s.%s)   : %7d",
                     $prod, $::FileExtNorm, $lNrm);
        if ($lNrm == $lpackNrmSum + $lRen - $lAdd) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        $pack = $::FileExtArchiveFiles;
        printf(CHECK " = normal. Log for archives (%s.%s) : %7d%s",
                     $prod.$pack, $::FileExtNorm, $lNrmP{$pack}, $::EOL);
        printf(CHECK " - added lines in archive log (%s.%s)  : %7d%s",
                     $prod, $::FileExtPackedAdd, $lAdd, $::EOL);
        foreach $pack (sort keys %{$::listsAll{$tb}->{"packed"}->{"prg"}}) {
            printf(CHECK " + normalized Log for %s (%s.%s)   : %7d%s",
                         $pack, $prod.$pack, $::FileExtNorm, $lNrmP{$pack},
                         $::EOL);
        }
        printf(CHECK " + remaining lines (%s.%s)             : %7d",
                     $prod, $::FileExtNormRemPacked, $lRen);
        if ($lRen != 0) { print(CHECK "    Key ERROR!!!"); }
        print(CHECK $::EOL);

        print(CHECK "============================================",
                    "=============================", $::EOL);
        printf(CHECK "   files scanned but ok      (%s.%s)   : %7d",
                     $prod, $::FileExtScannedOK, $lSbo);
        if ($lSbo == $lpackSboSum + $lRps - $lAdo) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        $pack = $::FileExtArchiveFiles;
        printf(CHECK " = 'sbo'-files for archives (%s.%s) : %7d%s",
                     $prod.$pack, $::FileExtScannedOK, $lSboP{$pack}, $::EOL);
        printf(CHECK " - added lines (%s.%s)                 : %7d%s",
                     $prod, $::FileExtPackedOKAdd, $lAdo, $::EOL);
        foreach $pack (sort keys %{$::listsAll{$tb}->{"packed"}->{"prg"}}) {
            printf(CHECK " + 'sbo'-files for %s (%s.%s)      : %7d%s",
                         $pack, $prod.$pack, $::FileExtScannedOK,
                         $lSboP{$pack}, $::EOL);
        }
        printf(CHECK " + remaining lines (%s.%s)             : %7d",
                     $prod, $::FileExtScanOKRemPacked, $lRps);
        if ($lRps != 0) { print(CHECK "    Key ERROR!!!"); }

        print(CHECK $::EOL);
        print(CHECK "============================================",
                    "=============================", $::EOL, $::EOL, $::EOL);
        close(CHECK);

        # append all other controlfiles to the main controlfile
        $pack = $::FileExtArchiveFiles;
        my $appFile = $prod . $pack  . "." . $::FileExtControl;
        ::appendFile($checkFile, $appFile) or do {
            ::log("ERROR appending $appFile to $checkFile: $!\n", 2);
        };

        foreach $pack (sort keys %{$::listsAll{$tb}->{"packed"}->{"prg"}}) {
            $appFile = $prod . $pack  . "." . $::FileExtControl;
            ::appendFile($checkFile, $appFile) or do {
                ::log("ERROR appending $appFile to $checkFile: $!\n", 2);
            }
        }

    }
}


# check evaluation list evaluation :-)
sub checkList {
    my $tb = $_[0];
    my $evalList = $_[1];
    my $prodEval = $_[2];
    my $checkFile = $_[3];
    my $nrViruses = $_[4];
    my $nrFiles = $_[5];
    my ($lM_i, $lM_s, $lAno, $lEv1, $lDet, $lViruses,
        $nrVirusesFound, $nrFilesFound, $percVirusesFound, $percFilesFound);
    # get check data
    ($lViruses, undef) = ::wc($::ListDir.$::sep.$evalList);
    ($lEv1, undef) = ::wc($prodEval . "." . $::FileExtEval);
    ($lM_i, undef) = ::wc($prodEval . "." . $::FileExtMultInf);
    ($lM_s, undef) = ::wc($prodEval . "." . $::FileExtMultScan);
    ($lAno, undef) = ::wc($prodEval . "." . $::FileExtFilteredAno);
    # only count lines with detected viruses (= field 2)
    ($lDet, undef) = ::wcSelect($prodEval . "." . $::FileExtDetVir, 2);

    my @datArray = ::getDataArray($prodEval.".".$::FileExtResult);
    $nrVirusesFound = $datArray[1];
    $percVirusesFound = $datArray[2];
    $nrFilesFound = $datArray[7];
    $percFilesFound = $datArray[8];
    if ($::debug > 1) {
        ::log("datArray = @datArray\n", 4, "DEBUG: ");
    }

    # write check data
    my $openError = 0;
    open(CHECK, ">> $checkFile") or do {
        ::log("ERROR opening $checkFile: $!\n", 2);
        ::log("You must check this manually!\n", 5);
        $openError = 1;
    };
    binmode(CHECK);

    if (not $openError) {
        print(CHECK "---- checks for evaluation list: ",$evalList,
                    " -------------------------------", $::EOL);
        printf(CHECK "   files found by product (%s.%s)      : %7d%s",
                     $prodEval, $::FileExtEval, $lEv1, $::EOL);
        printf(CHECK " - multiple infected files (%s.%s)     : %7d%s",
                     $prodEval, $::FileExtMultInf, $lM_i, $::EOL);
        printf(CHECK " - multiple scanned files (%s.%s)      : %7d%s",
                     $prodEval, $::FileExtMultScan, $lM_s, $::EOL);
        printf(CHECK " - anomaly files (%s.%s)               : %7d%s",
                     $prodEval, $::FileExtFilteredAno, $lAno, $::EOL);
        printf(CHECK " = infected files found (from %s.%s)   : %7d",
                     $prodEval, $::FileExtResult, $nrFilesFound);
        if (($lEv1 - $lM_i - $lM_s - $lAno ) == $nrFilesFound) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "   infected directories found (%s.%s)  : %7d%s",
                     $prodEval, $::FileExtDetVir, $lDet, $::EOL);
        printf(CHECK " = viruses found (from %s.%s)          : %7d",
                     $prodEval, $::FileExtResult, $nrVirusesFound);
        if ($lDet == $nrVirusesFound) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL, $::EOL);

        print(CHECK "---- result data for: ",$evalList,
                    " ------------------------------------------", $::EOL);
        my $spaceFill = 8 - length($prodEval);
        printf(CHECK "Infected files found (from %s.%3s)      ".
                     " "x$spaceFill . ": %7d  %5s%%%s",
                     $prodEval, $::FileExtResult,
                     $nrFilesFound, $percFilesFound, $::EOL);
        printf(CHECK "Infected files in testbed (from %8s.%3s) :".
                     " %7d  100.0%%%s",
                     $tb, $::FileExtSummary, $nrFiles, $::EOL);
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL);
        printf(CHECK "Malware/Viruses found (from %s.%3s)     ".
                     " "x$spaceFill . ": %7d  %5s%%%s",
                     $prodEval, $::FileExtResult,
                     $nrVirusesFound, $percVirusesFound, $::EOL);
        printf(CHECK "Malware/Viruses in testbed (from %8s.%3s):".
                     " %7d  100.0%%%s",
                     $tb, $::FileExtSummary, $nrViruses, $::EOL);
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL, $::EOL, $::EOL);
    }
    close(CHECK);
}


# check 'false positve' evaluation list evaluation :-)
sub checkListFP {
    my $tb = $_[0];
    my $evalList = $_[1];
    my $prodEval = $_[2];
    my $checkFile = $_[3];
    my $nrViruses = $_[4];  # not used as it makes no sense to group fp samples
    my $nrFiles = $_[5];    # number of false positve samples in the testbed

    my ($lM_i, $lM_s, $lAno, $lFpEv1, $nrFilesFound, $percFilesFound);
    # get check data
    ($lFpEv1, undef) = ::wc($prodEval . "." . $::FileExtEval);
    ($lM_i, undef) = ::wc($prodEval . "." . $::FileExtMultInf);
    ($lM_s, undef) = ::wc($prodEval . "." . $::FileExtMultScan);
    ($lAno, undef) = ::wc($prodEval . "." . $::FileExtFilteredAno);

    my @datArray = ::getDataArray($prodEval.".".$::FileExtResult);
    $nrFilesFound = $datArray[1];
    $percFilesFound = $datArray[2];
    if ($::debug > 1) {
        ::log("datArray = @datArray\n", 4, "DEBUG: ");
    }

    # write check data
    my $openError = 0;
    open(CHECK, ">> $checkFile") or do {
        ::log("ERROR opening $checkFile: $!\n", 2);
        ::log("You must check this manually!\n", 5);
        $openError = 1;
    };
    binmode(CHECK);

    if (not $openError) {
        print(CHECK "---- checks for evaluation list: ",$evalList,
                    " -------------------------------", $::EOL);
        printf(CHECK "   false positves reported as infected (%s.%s): %7d%s",
                     $prodEval, $::FileExtEval, $lFpEv1, $::EOL);
        printf(CHECK " - multiple infected false positves (%s.%s)   : %7d%s",
                     $prodEval, $::FileExtMultInf, $lM_i, $::EOL);
        printf(CHECK " - multiple scanned false positves (%s.%s)    : %7d%s",
                     $prodEval, $::FileExtMultScan, $lM_s, $::EOL);
        printf(CHECK " - anomaly files (%s.%s)                      : %7d%s",
                     $prodEval, $::FileExtFilteredAno, $lAno, $::EOL);
        printf(CHECK " = False positives found (from %s.%s)         : %7d",
                     $prodEval, $::FileExtResult, $nrFilesFound);
        if (($lFpEv1 - $lM_i - $lM_s - $lAno) == $nrFilesFound) {
            print(CHECK "    ok", $::EOL);
        } else {
            print(CHECK "    NOT OK!!!", $::EOL);
        }
        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL, $::EOL);
        print(CHECK "---- result data for: ", $evalList,
                    " ------------------------------------------", $::EOL);
        my $spaceFill = length($prodEval);
        printf(CHECK "False positives reported as infected (from %s.%3s):".
                     " %7d  %5s%%%s", $prodEval, $::FileExtResult,
                     $nrFilesFound, $percFilesFound, $::EOL);
        printf(CHECK "False positives in testbed (from %8s.%3s)  ".
                     " "x$spaceFill . ": %7d  100%%%s",
                     $tb, $::FileExtSummary, $nrFiles, $::EOL);

        print(CHECK "--------------------------------------------",
                    "-----------------------------", $::EOL, $::EOL, $::EOL);
    }
    close(CHECK);
}



# return a true value
1;